@@ -2,6 +2,7 @@
 #include <bpf/libbpf.h>
 #include <cstdio>
 #include <cstring>
+#include <ctime>
 #include <fstream>
 #include <getopt.h>
 #include <iostream>
@@ -9,7 +10,6 @@
 #include <optional>
 #include <sys/resource.h>
 #include <sys/utsname.h>
-#include <time.h>
 #include <unistd.h>
 
 #include "aot/aot.h"
@@ -18,7 +18,6 @@
 #include "ast/passes/codegen_llvm.h"
 #include "ast/passes/config_analyser.h"
 #include "ast/passes/field_analyser.h"
-#include "ast/passes/node_counter.h"
 #include "ast/passes/portability_analyser.h"
 #include "ast/passes/resource_analyser.h"
 #include "ast/passes/return_path_analyser.h"
@@ -78,73 +77,76 @@ enum Options {
   EMIT_ELF,
   EMIT_LLVM,
   NO_FEATURE,
+  DEBUG,
+  DRY_RUN,
 };
 } // namespace
 
-void usage()
+void usage(std::ostream& out)
 {
   // clang-format off
-  std::cerr << "USAGE:" << std::endl;
-  std::cerr << "    bpftrace [options] filename" << std::endl;
-  std::cerr << "    bpftrace [options] - <stdin input>" << std::endl;
-  std::cerr << "    bpftrace [options] -e 'program'" << std::endl;
-  std::cerr << std::endl;
-  std::cerr << "OPTIONS:" << std::endl;
-  std::cerr << "    -B MODE        output buffering mode ('full', 'none')" << std::endl;
-  std::cerr << "    -f FORMAT      output format ('text', 'json')" << std::endl;
-  std::cerr << "    -o file        redirect bpftrace output to file" << std::endl;
-  std::cerr << "    -e 'program'   execute this program" << std::endl;
-  std::cerr << "    -h, --help     show this help message" << std::endl;
-  std::cerr << "    -I DIR         add the directory to the include search path" << std::endl;
-  std::cerr << "    --include FILE add an #include file before preprocessing" << std::endl;
-  std::cerr << "    -l [search|filename]" << std::endl;
-  std::cerr << "                   list kernel probes or probes in a program" << std::endl;
-  std::cerr << "    -p PID         enable USDT probes on PID" << std::endl;
-  std::cerr << "    -c 'CMD'       run CMD and enable USDT probes on resulting process" << std::endl;
-  std::cerr << "    --usdt-file-activation" << std::endl;
-  std::cerr << "                   activate usdt semaphores based on file path" << std::endl;
-  std::cerr << "    --unsafe       allow unsafe builtin functions" << std::endl;
-  std::cerr << "    -q             keep messages quiet" << std::endl;
-  std::cerr << "    --info         Print information about kernel BPF support" << std::endl;
-  std::cerr << "    -k             emit a warning when a bpf helper returns an error (except read functions)" << std::endl;
-  std::cerr << "    -kk            check all bpf helper functions" << std::endl;
-  std::cerr << "    -V, --version  bpftrace version" << std::endl;
-  std::cerr << "    --no-warnings  disable all warning messages" << std::endl;
-  std::cerr << std::endl;
-  std::cerr << "TROUBLESHOOTING OPTIONS:" << std::endl;
-  std::cerr << "    -v                      verbose messages" << std::endl;
-  std::cerr << "    -d                      (dry run) debug info" << std::endl;
-  std::cerr << "    -dd                     (dry run) verbose debug info" << std::endl;
-  std::cerr << "    --emit-elf FILE         (dry run) generate ELF file with bpf programs and write to FILE" << std::endl;
-  std::cerr << "    --emit-llvm FILE        write LLVM IR to FILE.original.ll and FILE.optimized.ll" << std::endl;
-  std::cerr << std::endl;
-  std::cerr << "ENVIRONMENT:" << std::endl;
-  std::cerr << "    BPFTRACE_BTF                      [default: none] BTF file" << std::endl;
-  std::cerr << "    BPFTRACE_CACHE_USER_SYMBOLS       [default: auto] enable user symbol cache" << std::endl;
-  std::cerr << "    BPFTRACE_CPP_DEMANGLE             [default: 1] enable C++ symbol demangling" << std::endl;
-  std::cerr << "    BPFTRACE_DEBUG_OUTPUT             [default: 0] enable bpftrace's internal debugging outputs" << std::endl;
-  std::cerr << "    BPFTRACE_KERNEL_BUILD             [default: /lib/modules/$(uname -r)] kernel build directory" << std::endl;
-  std::cerr << "    BPFTRACE_KERNEL_SOURCE            [default: /lib/modules/$(uname -r)] kernel headers directory" << std::endl;
-  std::cerr << "    BPFTRACE_LAZY_SYMBOLICATION       [default: 0] symbolicate lazily/on-demand" << std::endl;
-  std::cerr << "    BPFTRACE_LOG_SIZE                 [default: 1000000] log size in bytes" << std::endl;
-  std::cerr << "    BPFTRACE_MAX_BPF_PROGS            [default: 512] max number of generated BPF programs" << std::endl;
-  std::cerr << "    BPFTRACE_MAX_CAT_BYTES            [default: 10k] maximum bytes read by cat builtin" << std::endl;
-  std::cerr << "    BPFTRACE_MAX_MAP_KEYS             [default: 4096] max keys in a map" << std::endl;
-  std::cerr << "    BPFTRACE_MAX_PROBES               [default: 512] max number of probes" << std::endl;
-  std::cerr << "    BPFTRACE_MAX_STRLEN               [default: 64] bytes on BPF stack per str()" << std::endl;
-  std::cerr << "    BPFTRACE_MAX_TYPE_RES_ITERATIONS  [default: 0] number of levels of nested field accesses for tracepoint args" << std::endl;
-  std::cerr << "    BPFTRACE_PERF_RB_PAGES            [default: 64] pages per CPU to allocate for ring buffer" << std::endl;
-  std::cerr << "    BPFTRACE_STACK_MODE               [default: bpftrace] Output format for ustack and kstack builtins" << std::endl;
-  std::cerr << "    BPFTRACE_STR_TRUNC_TRAILER        [default: '..'] string truncation trailer" << std::endl;
-  std::cerr << "    BPFTRACE_VMLINUX                  [default: none] vmlinux path used for kernel symbol resolution" << std::endl;
-  std::cerr << std::endl;
-  std::cerr << "EXAMPLES:" << std::endl;
-  std::cerr << "bpftrace -l '*sleep*'" << std::endl;
-  std::cerr << "    list probes containing \"sleep\"" << std::endl;
-  std::cerr << "bpftrace -e 'kprobe:do_nanosleep { printf(\"PID %d sleeping...\\n\", pid); }'" << std::endl;
-  std::cerr << "    trace processes calling sleep" << std::endl;
-  std::cerr << "bpftrace -e 'tracepoint:raw_syscalls:sys_enter { @[comm] = count(); }'" << std::endl;
-  std::cerr << "    count syscalls by process name" << std::endl;
+  out << "USAGE:" << std::endl;
+  out << "    bpftrace [options] filename" << std::endl;
+  out << "    bpftrace [options] - <stdin input>" << std::endl;
+  out << "    bpftrace [options] -e 'program'" << std::endl;
+  out << std::endl;
+  out << "OPTIONS:" << std::endl;
+  out << "    -B MODE        output buffering mode ('line', 'full', 'none')" << std::endl;
+  out << "    -f FORMAT      output format ('text', 'json')" << std::endl;
+  out << "    -o file        redirect bpftrace output to file" << std::endl;
+  out << "    -e 'program'   execute this program" << std::endl;
+  out << "    -h, --help     show this help message" << std::endl;
+  out << "    -I DIR         add the directory to the include search path" << std::endl;
+  out << "    --include FILE add an #include file before preprocessing" << std::endl;
+  out << "    -l [search|filename]" << std::endl;
+  out << "                   list kernel probes or probes in a program" << std::endl;
+  out << "    -p PID         enable USDT probes on PID" << std::endl;
+  out << "    -c 'CMD'       run CMD and enable USDT probes on resulting process" << std::endl;
+  out << "    --usdt-file-activation" << std::endl;
+  out << "                   activate usdt semaphores based on file path" << std::endl;
+  out << "    --unsafe       allow unsafe/destructive functionality" << std::endl;
+  out << "    -q             keep messages quiet" << std::endl;
+  out << "    --info         Print information about kernel BPF support" << std::endl;
+  out << "    -k             emit a warning when probe read helpers return an error" << std::endl;
+  out << "    -V, --version  bpftrace version" << std::endl;
+  out << "    --no-warnings  disable all warning messages" << std::endl;
+  out << std::endl;
+  out << "TROUBLESHOOTING OPTIONS:" << std::endl;
+  out << "    -v                      verbose messages" << std::endl;
+  out << "    --dry-run               terminate execution right after attaching all the probes" << std::endl;
+  out << "    -d STAGE                debug info for various stages of bpftrace execution" << std::endl;
+  out << "                            ('all', 'ast', 'codegen', 'codegen-opt', 'dis', 'libbpf', 'verifier')" << std::endl;
+  out << "    --emit-elf FILE         (dry run) generate ELF file with bpf programs and write to FILE" << std::endl;
+  out << "    --emit-llvm FILE        write LLVM IR to FILE.original.ll and FILE.optimized.ll" << std::endl;
+  out << std::endl;
+  out << "ENVIRONMENT:" << std::endl;
+  out << "    BPFTRACE_BTF                      [default: none] BTF file" << std::endl;
+  out << "    BPFTRACE_CACHE_USER_SYMBOLS       [default: auto] enable user symbol cache" << std::endl;
+  out << "    BPFTRACE_COLOR                    [default: auto] enable log output colorization" << std::endl;
+  out << "    BPFTRACE_CPP_DEMANGLE             [default: 1] enable C++ symbol demangling" << std::endl;
+  out << "    BPFTRACE_DEBUG_OUTPUT             [default: 0] enable bpftrace's internal debugging outputs" << std::endl;
+  out << "    BPFTRACE_KERNEL_BUILD             [default: /lib/modules/$(uname -r)] kernel build directory" << std::endl;
+  out << "    BPFTRACE_KERNEL_SOURCE            [default: /lib/modules/$(uname -r)] kernel headers directory" << std::endl;
+  out << "    BPFTRACE_LAZY_SYMBOLICATION       [default: 0] symbolicate lazily/on-demand" << std::endl;
+  out << "    BPFTRACE_LOG_SIZE                 [default: 1000000] log size in bytes" << std::endl;
+  out << "    BPFTRACE_MAX_BPF_PROGS            [default: 1024] max number of generated BPF programs" << std::endl;
+  out << "    BPFTRACE_MAX_CAT_BYTES            [default: 10k] maximum bytes read by cat builtin" << std::endl;
+  out << "    BPFTRACE_MAX_MAP_KEYS             [default: 4096] max keys in a map" << std::endl;
+  out << "    BPFTRACE_MAX_PROBES               [default: 1024] max number of probes" << std::endl;
+  out << "    BPFTRACE_MAX_STRLEN               [default: 1024] bytes on BPF stack per str()" << std::endl;
+  out << "    BPFTRACE_MAX_TYPE_RES_ITERATIONS  [default: 0] number of levels of nested field accesses for tracepoint args" << std::endl;
+  out << "    BPFTRACE_PERF_RB_PAGES            [default: 64] pages per CPU to allocate for ring buffer" << std::endl;
+  out << "    BPFTRACE_STACK_MODE               [default: bpftrace] Output format for ustack and kstack builtins" << std::endl;
+  out << "    BPFTRACE_STR_TRUNC_TRAILER        [default: '..'] string truncation trailer" << std::endl;
+  out << "    BPFTRACE_VMLINUX                  [default: none] vmlinux path used for kernel symbol resolution" << std::endl;
+  out << std::endl;
+  out << "EXAMPLES:" << std::endl;
+  out << "bpftrace -l '*sleep*'" << std::endl;
+  out << "    list probes containing \"sleep\"" << std::endl;
+  out << R"(bpftrace -e 'kprobe:do_nanosleep { printf("PID %d sleeping...\n", pid); }')" << std::endl;
+  out << "    trace processes calling sleep" << std::endl;
+  out << "bpftrace -e 'tracepoint:raw_syscalls:sys_enter { @[comm] = count(); }'" << std::endl;
+  out << "    count syscalls by process name" << std::endl;
   // clang-format on
 }
 
@@ -162,29 +164,21 @@ static void enforce_infinite_rlimit()
                  << "\"ulimit -l 8192\" to fix the problem";
 }
 
-void check_is_root()
-{
-  if (geteuid() != 0) {
-    LOG(ERROR) << "bpftrace currently only supports running as the root user.";
-    exit(1);
-  }
-}
-
 static void info(BPFnofeature no_feature)
 {
   struct utsname utsname;
   uname(&utsname);
 
-  std::cerr << "System" << std::endl
+  std::cout << "System" << std::endl
             << "  OS: " << utsname.sysname << " " << utsname.release << " "
             << utsname.version << std::endl
             << "  Arch: " << utsname.machine << std::endl;
 
-  std::cerr << std::endl;
-  std::cerr << BuildInfo::report();
+  std::cout << std::endl;
+  std::cout << BuildInfo::report();
 
-  std::cerr << std::endl;
-  std::cerr << BPFfeature(no_feature).report();
+  std::cout << std::endl;
+  std::cout << BPFfeature(no_feature).report();
 }
 
 static std::optional<struct timespec> get_delta_with_boottime(int clock_type)
@@ -317,7 +311,6 @@ static void parse_env(BPFtrace& bpftrace)
       exit(1);
   }
 
-  bpftrace.max_ast_nodes_ = std::numeric_limits<uint64_t>::max();
   get_uint64_env_var("BPFTRACE_MAX_AST_NODES",
                      [&](uint64_t x) { bpftrace.max_ast_nodes_ = x; });
 
@@ -343,9 +336,20 @@ static void parse_env(BPFtrace& bpftrace)
                     "BPFTRACE_MAX_MAP_KEYS instead.";
     config_setter.set(ConfigKeyInt::max_map_keys, x);
   });
+
+  get_bool_env_var("BPFTRACE_USE_BLAZESYM", [&](bool x) {
+#ifndef HAVE_BLAZESYM
+    if (x) {
+      LOG(ERROR) << "BPFTRACE_USE_BLAZESYM requires blazesym support enabled "
+                    "during build.";
+      exit(1);
+    }
+#endif
+    config_setter.set(ConfigKeyBool::use_blazesym, x);
+  });
 }
 
-[[nodiscard]] std::unique_ptr<ast::Program> parse(
+[[nodiscard]] std::optional<ast::ASTContext> parse(
     BPFtrace& bpftrace,
     const std::string& name,
     const std::string& program,
@@ -358,46 +362,39 @@ static void parse_env(BPFtrace& bpftrace)
 
   err = driver.parse();
   if (err)
-    return nullptr;
+    return std::nullopt;
 
   bpftrace.parse_btf(driver.list_modules());
 
-  ast::FieldAnalyser fields(driver.root.get(), bpftrace);
+  ast::FieldAnalyser fields(driver.ctx, bpftrace);
   err = fields.analyse();
   if (err)
-    return nullptr;
+    return std::nullopt;
 
-  if (TracepointFormatParser::parse(driver.root.get(), bpftrace) == false)
-    return nullptr;
+  if (TracepointFormatParser::parse(driver.ctx, bpftrace) == false)
+    return std::nullopt;
 
   // NOTE(mmarchini): if there are no C definitions, clang parser won't run to
   // avoid issues in some versions. Since we're including files in the command
   // line, we want to force parsing, so we make sure C definitions are not
   // empty before going to clang parser stage.
-  if (!include_files.empty() && driver.root->c_definitions.empty())
-    driver.root->c_definitions = "#define __BPFTRACE_DUMMY__";
+  if (!include_files.empty() && driver.ctx.root->c_definitions.empty())
+    driver.ctx.root->c_definitions = "#define __BPFTRACE_DUMMY__";
 
-  bool should_clang_parse = !(driver.root.get()->c_definitions.empty() &&
+  bool should_clang_parse = !(driver.ctx.root->c_definitions.empty() &&
                               bpftrace.btf_set_.empty());
 
   if (should_clang_parse) {
     ClangParser clang;
+    std::string ksrc, kobj;
+    struct utsname utsname;
     std::vector<std::string> extra_flags;
-    {
-      struct utsname utsname;
-      uname(&utsname);
-      std::string ksrc, kobj;
-      auto kdirs = get_kernel_dirs(utsname);
-      ksrc = std::get<0>(kdirs);
-      kobj = std::get<1>(kdirs);
+    uname(&utsname);
+    bool found_kernel_headers = get_kernel_dirs(utsname, ksrc, kobj);
 
-      if (ksrc != "") {
-        extra_flags = get_kernel_cflags(
-            utsname.machine, ksrc, kobj, bpftrace.kconfig);
-      }
-    }
-    extra_flags.push_back("-include");
-    extra_flags.push_back("/bpftrace/include/" CLANG_WORKAROUNDS_H);
+    if (found_kernel_headers)
+      extra_flags = get_kernel_cflags(
+          utsname.machine, ksrc, kobj, bpftrace.kconfig);
 
     for (auto dir : include_dirs) {
       extra_flags.push_back("-I");
@@ -408,15 +405,27 @@ static void parse_env(BPFtrace& bpftrace)
       extra_flags.push_back(file);
     }
 
-    if (!clang.parse(driver.root.get(), bpftrace, extra_flags))
-      return nullptr;
+    if (!clang.parse(driver.ctx.root, bpftrace, extra_flags)) {
+      if (!found_kernel_headers) {
+        LOG(WARNING)
+            << "Could not find kernel headers in " << ksrc << " / " << kobj
+            << ". To specify a particular path to kernel headers, set the env "
+            << "variables BPFTRACE_KERNEL_SOURCE and, optionally, "
+            << "BPFTRACE_KERNEL_BUILD if the kernel was built in a different "
+            << "directory than its source. You can also point the variable to "
+            << "a directory with built-in headers extracted from the following "
+            << "snippet:\nmodprobe kheaders && tar -C <directory> -xf "
+            << "/sys/kernel/kheaders.tar.xz";
+      }
+      return {};
+    }
   }
 
   err = driver.parse();
   if (err)
-    return nullptr;
+    return {};
 
-  return std::move(driver.root);
+  return std::move(driver.ctx);
 }
 
 ast::PassManager CreateDynamicPM()
@@ -424,7 +433,6 @@ ast::PassManager CreateDynamicPM()
   ast::PassManager pm;
   pm.AddPass(ast::CreateConfigPass());
   pm.AddPass(ast::CreateSemanticPass());
-  pm.AddPass(ast::CreateCounterPass());
   pm.AddPass(ast::CreateResourcePass());
   pm.AddPass(ast::CreateReturnPathPass());
 
@@ -448,7 +456,8 @@ struct Args {
   bool listing = false;
   bool safe_mode = true;
   bool usdt_file_activation = false;
-  int helper_check_level = 0;
+  int helper_check_level = 1;
+  bool no_warnings = false;
   TestMode test_mode = TestMode::UNSET;
   std::string script;
   std::string search;
@@ -464,13 +473,33 @@ struct Args {
   std::vector<std::string> include_dirs;
   std::vector<std::string> include_files;
   std::vector<std::string> params;
+  std::vector<std::string> debug_stages;
 };
 
+static bool parse_debug_stages(const std::string& arg)
+{
+  auto stages = split_string(arg, ',', /* remove_empty= */ true);
+
+  for (const auto& stage : stages) {
+    if (debug_stages.contains(stage)) {
+      bt_debug.insert(debug_stages.at(stage));
+    } else if (stage == "all") {
+      for (const auto& [_, s] : debug_stages)
+        bt_debug.insert(s);
+    } else {
+      LOG(ERROR) << "USAGE: invalid option for -d: " << stage;
+      return false;
+    }
+  }
+
+  return true;
+}
+
 Args parse_args(int argc, char* argv[])
 {
   Args args;
 
-  const char* const short_options = "dbB:f:e:hlp:vqc:Vo:I:k";
+  const char* const short_options = "d:bB:f:e:hlp:vqc:Vo:I:k";
   option long_options[] = {
     option{ "help", no_argument, nullptr, Options::HELP },
     option{ "version", no_argument, nullptr, Options::VERSION },
@@ -486,10 +515,13 @@ Args parse_args(int argc, char* argv[])
     option{ "test", required_argument, nullptr, Options::TEST },
     option{ "aot", required_argument, nullptr, Options::AOT },
     option{ "no-feature", required_argument, nullptr, Options::NO_FEATURE },
+    option{ "debug", required_argument, nullptr, Options::DEBUG },
+    option{ "dry-run", no_argument, nullptr, Options::DRY_RUN },
     option{ nullptr, 0, nullptr, 0 }, // Must be last
   };
 
   int c;
+  bool has_k = false;
   while ((c = getopt_long(argc, argv, short_options, long_options, nullptr)) !=
          -1) {
     switch (c) {
@@ -506,6 +538,8 @@ Args parse_args(int argc, char* argv[])
         break;
       case Options::NO_WARNING: // --no-warnings
         DISABLE_LOG(WARNING);
+        args.no_warnings = true;
+        args.helper_check_level = 0;
         break;
       case Options::TEST: // --test
         if (std::strcmp(optarg, "codegen") == 0)
@@ -526,15 +560,16 @@ Args parse_args(int argc, char* argv[])
           exit(1);
         }
         break;
+      case Options::DRY_RUN:
+        dry_run = true;
+        break;
       case 'o':
         args.output_file = optarg;
         break;
       case 'd':
-        bt_debug++;
-        if (bt_debug == DebugLevel::kNone) {
-          usage();
+      case Options::DEBUG:
+        if (!parse_debug_stages(optarg))
           exit(1);
-        }
         break;
       case 'q':
         bt_quiet = true;
@@ -587,45 +622,45 @@ Args parse_args(int argc, char* argv[])
         break;
       case 'h':
       case Options::HELP:
-        usage();
+        usage(std::cout);
         exit(0);
       case 'V':
       case Options::VERSION:
         std::cout << "bpftrace " << BPFTRACE_VERSION << std::endl;
         exit(0);
       case 'k':
-        args.helper_check_level++;
-        if (args.helper_check_level >= 3) {
-          usage();
+        if (has_k) {
+          LOG(ERROR) << "USAGE: -kk has been deprecated. Use a single -k for "
+                        "runtime warnings for errors in map "
+                        "lookups and probe reads.";
           exit(1);
         }
+        if (!args.no_warnings) {
+          args.helper_check_level = 2;
+        }
+        has_k = true;
         break;
       default:
-        usage();
+        usage(std::cerr);
         exit(1);
     }
   }
 
   if (argc == 1) {
-    usage();
-    exit(1);
-  }
-
-  if (bt_verbose && (bt_debug != DebugLevel::kNone)) {
-    // TODO: allow both
-    LOG(ERROR) << "USAGE: Use either -v or -d.";
+    usage(std::cerr);
     exit(1);
   }
 
   if (!args.cmd_str.empty() && !args.pid_str.empty()) {
     LOG(ERROR) << "USAGE: Cannot use both -c and -p.";
-    usage();
+    usage(std::cerr);
     exit(1);
   }
 
   // Difficult to serialize flex generated types
-  if (args.helper_check_level && args.build_mode == BuildMode::AHEAD_OF_TIME) {
-    LOG(ERROR) << "Cannot use -k[k] with --aot";
+  if (args.helper_check_level == 2 &&
+      args.build_mode == BuildMode::AHEAD_OF_TIME) {
+    LOG(ERROR) << "Cannot use -k with --aot";
     exit(1);
   }
 
@@ -635,7 +670,7 @@ Args parse_args(int argc, char* argv[])
       args.search = "*:*";
     } else if (optind == argc - 1) {
       std::string_view val(argv[optind]);
-      if (std_filesystem::exists(val)) {
+      if (std::filesystem::exists(val)) {
         args.filename = val;
       } else {
         if (val == "*") {
@@ -646,7 +681,7 @@ Args parse_args(int argc, char* argv[])
       }
       optind++;
     } else {
-      usage();
+      usage(std::cerr);
       exit(1);
     }
   } else {
@@ -673,12 +708,32 @@ Args parse_args(int argc, char* argv[])
   return args;
 }
 
+bool is_colorize()
+{
+  const char* color_env = std::getenv("BPFTRACE_COLOR");
+  if (!color_env) {
+    return isatty(STDOUT_FILENO) && isatty(STDERR_FILENO);
+  }
+
+  std::string_view mode(color_env);
+  if (mode == "always") {
+    return true;
+  } else if (mode == "never") {
+    return false;
+  } else {
+    if (mode != "auto") {
+      LOG(WARNING) << "Invalid env value! The valid values of `BPFTRACE_COLOR` "
+                      "are [auto|always|never]. The current value is "
+                   << mode << "!";
+    }
+    return isatty(STDOUT_FILENO) && isatty(STDERR_FILENO);
+  }
+}
+
 int main(int argc, char* argv[])
 {
-  int err;
-
+  Log::get().set_colorize(is_colorize());
   const Args args = parse_args(argc, argv);
-
   std::ostream* os = &std::cout;
   std::ofstream outputstream;
   if (!args.output_file.empty()) {
@@ -705,24 +760,21 @@ int main(int argc, char* argv[])
   switch (args.obc) {
     case OutputBufferConfig::UNSET:
     case OutputBufferConfig::LINE:
-      std::setvbuf(stdout, NULL, _IOLBF, BUFSIZ);
+      std::setvbuf(stdout, nullptr, _IOLBF, BUFSIZ);
       break;
     case OutputBufferConfig::FULL:
-      std::setvbuf(stdout, NULL, _IOFBF, BUFSIZ);
+      std::setvbuf(stdout, nullptr, _IOFBF, BUFSIZ);
       break;
     case OutputBufferConfig::NONE:
-      std::setvbuf(stdout, NULL, _IONBF, BUFSIZ);
+      std::setvbuf(stdout, nullptr, _IONBF, BUFSIZ);
       break;
   }
 
   libbpf_set_print(libbpf_print);
 
-  Config config = Config(!args.cmd_str.empty(), bt_verbose);
+  Config config = Config(!args.cmd_str.empty());
   BPFtrace bpftrace(std::move(output), args.no_feature, config);
 
-  if (!args.cmd_str.empty())
-    bpftrace.cmd_ = args.cmd_str;
-
   parse_env(bpftrace);
 
   bpftrace.usdt_file_activation_ = args.usdt_file_activation;
@@ -785,12 +837,12 @@ int main(int argc, char* argv[])
 
     bpftrace.parse_btf(driver.list_modules());
 
-    ast::SemanticAnalyser semantics(driver.root.get(), bpftrace, false, true);
+    ast::SemanticAnalyser semantics(driver.ctx, bpftrace, false, true);
     err = semantics.analyse();
     if (err)
       return err;
 
-    bpftrace.probe_matcher_->list_probes(driver.root.get());
+    bpftrace.probe_matcher_->list_probes(driver.ctx.root);
     return 0;
   }
 
@@ -846,59 +898,47 @@ int main(int argc, char* argv[])
   // rlimit?
   enforce_infinite_rlimit();
 
-  auto ast_prog = parse(
+  auto ast_ctx = parse(
       bpftrace, filename, program, args.include_dirs, args.include_files);
-  if (!ast_prog)
+  if (!ast_ctx)
     return 1;
 
   if (args.listing) {
-    bpftrace.probe_matcher_->list_probes(ast_prog.get());
+    bpftrace.probe_matcher_->list_probes(ast_ctx->root);
     return 0;
   }
 
-  ast::PassContext ctx(bpftrace);
+  ast::PassContext ctx(bpftrace, *ast_ctx);
   ast::PassManager pm;
   switch (args.build_mode) {
     case BuildMode::DYNAMIC:
       pm = CreateDynamicPM();
       break;
     case BuildMode::AHEAD_OF_TIME:
+      if (bpftrace.has_dwarf_data()) {
+        // See #3392 to learn why AOT does not yet support uprobe+DebugInfo.
+        LOG(ERROR) << "AOT does not yet support uprobe probes using DebugInfo.";
+        if (std::getenv("__BPFTRACE_NOTIFY_AOT_PORTABILITY_DISABLED"))
+          std::cout << "__BPFTRACE_NOTIFY_AOT_PORTABILITY_DISABLED"
+                    << std::endl;
+      }
       pm = CreateAotPM();
       break;
   }
 
-  bpftrace.kfunc_recursion_check(ast_prog.get());
+  bpftrace.fentry_recursion_check(ast_ctx->root);
 
-  auto pmresult = pm.Run(std::move(ast_prog), ctx);
+  auto pmresult = pm.Run(ctx);
   if (!pmresult.Ok())
     return 1;
 
-  auto ast_root = std::unique_ptr<ast::Node>(pmresult.Root());
-
-  if (!bpftrace.cmd_.empty()) {
-    try {
-      bpftrace.child_ = std::make_unique<ChildProc>(args.cmd_str);
-    } catch (const std::runtime_error& e) {
-      LOG(ERROR) << "Failed to fork child: " << e.what();
-      exit(1);
-    }
-  }
-
-  err = bpftrace.create_pcaps();
-  if (err) {
-    LOG(ERROR) << "Failed to create pcap file";
-    return err;
-  }
-
-  ast::CodegenLLVM llvm(&*ast_root,
-                        bpftrace,
-                        args.build_mode == BuildMode::AHEAD_OF_TIME);
+  ast::CodegenLLVM llvm(ctx.ast_ctx, bpftrace);
   BpfBytecode bytecode;
   try {
     llvm.generate_ir();
-    if (bt_debug == DebugLevel::kFullDebug) {
-      std::cout << "Before optimization\n";
-      std::cout << "-------------------\n\n";
+    if (bt_debug.find(DebugStage::Codegen) != bt_debug.end()) {
+      std::cout << "LLVM IR before optimization\n";
+      std::cout << "---------------------------\n\n";
       llvm.DumpIR();
     }
     if (!args.output_llvm.empty()) {
@@ -914,11 +954,9 @@ int main(int argc, char* argv[])
     }
 
     llvm.optimize();
-    if (bt_debug != DebugLevel::kNone) {
-      if (bt_debug == DebugLevel::kFullDebug) {
-        std::cout << "\nAfter optimization\n";
-        std::cout << "------------------\n\n";
-      }
+    if (bt_debug.find(DebugStage::CodegenOpt) != bt_debug.end()) {
+      std::cout << "\nLLVM IR after optimization\n";
+      std::cout << "----------------------------\n\n";
       llvm.DumpIR();
     }
     if (!args.output_llvm.empty()) {
@@ -936,7 +974,9 @@ int main(int argc, char* argv[])
       return aot::generate(
           bpftrace.resources, args.aot, aot_output.data(), aot_output.size());
     }
-    bytecode = llvm.emit();
+
+    bool disassemble = bt_debug.find(DebugStage::Disassemble) != bt_debug.end();
+    bytecode = llvm.emit(disassemble);
   } catch (const std::system_error& ex) {
     LOG(ERROR) << "failed to write elf: " << ex.what();
     return 1;
@@ -945,7 +985,7 @@ int main(int argc, char* argv[])
     return 1;
   }
 
-  if (bt_debug != DebugLevel::kNone || args.test_mode == TestMode::CODEGEN)
+  if (args.test_mode == TestMode::CODEGEN)
     return 0;
 
   return run_bpftrace(bpftrace, bytecode);
